/*
 * linux/include/asm-arm/arch-adifcc/timex.h
 *
 * XScale architecture timex specifications
 */

/* This is for a timer based on the XS80200's PMU counter */

#define CLOCK_TICK_RATE 600000000 /* Underlying HZ */

