/*
 * TQM8260 board specific definitions
 * 
 * Copyright (c) 2001 Wolfgang Denk (wd@denx.de)
 */

#ifndef __MACH_TQM8260_H
#define __MACH_TQM8260_H

#include <linux/config.h>
 
#include <asm/ppcboot.h>

#define IMAP_ADDR		((uint)0xFFF00000)
#define PHY_INTERRUPT		25

#endif	/* __MACH_TQM8260_H */
