#define MACHVEC_PLATFORM_NAME	sn1
#include <asm/machvec_init.h>
#include <asm/io.h>
#include <linux/pci.h>
void*
sn1_mk_io_addr_MACRO

dma_addr_t
sn1_pci_map_single_MACRO

int
sn1_pci_map_sg_MACRO

unsigned long
sn1_virt_to_phys_MACRO

void *
sn1_phys_to_virt_MACRO
