/*
 *  linux/include/asm-arm/hdreg.h
 *
 *  Copyright (C) 1994-1996  Linus Torvalds & authors
 */

#ifndef __ASMS390_HDREG_H
#define __ASMS390_HDREG_H

typedef unsigned long ide_ioreg_t;

#endif /* __ASMS390_HDREG_H */

