/*
 * linux/include/asm-arm/arch-cl7500/timex.h
 *
 * CL7500 architecture timex specifications
 *
 * Copyright (C) 1999 Nexus Electronics Ltd
 */

/*
 * On the ARM7500, the clock ticks at 2MHz.
 */
#define CLOCK_TICK_RATE		2000000

