/*
 *  linux/include/asm-arm/proc-armo/shmparam.h
 *
 *  Copyright (C) 1996 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  definitions for the shared process memory on the ARM3
 */
#ifndef __ASM_PROC_SHMPARAM_H
#define __ASM_PROC_SHMPARAM_H

#ifndef SHMMAX
#define SHMMAX		0x003fa000
#endif

#endif
