/*
 * linux/include/asm-arm/arch-80200fcc/dma.h
 *
 * Copyright (C) 2001 MontaVista Software, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#ifndef __ASM_ARCH_DMA_H
#define __ASM_ARCH_DMA_H

#define MAX_DMA_ADDRESS		0xffffffff

/* No DMA */
#define MAX_DMA_CHANNELS	0

#endif /* _ASM_ARCH_DMA_H */
