/*
 *  linux/include/asm-arm/arch-anakin/param.h
 *
 *  Copyright (C) 2001 Aleph One Ltd. for Acunia N.V.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Changelog:
 *   11-Apr-2001 TTC	Created
 */

/*
 * Reserved for future use
 */
